'''Game main module.

Contains the entry point used by the run_game.py script.

Feel free to put all your game code here, or in other modules in this "gamelib"
package.
'''

import data

from math import pi, sin, cos

import pyglet
from pyglet.gl import *

from pyglet.window import key

from terrain import Terrain

from world import World
from player import Player

# ==

# mutable global viewpoint

rx = ry = rz = 0

zoom = 1.0 # higher means camera is closer to target

rotating = False

# ==

# some toplevel code is from pyglet/examples/graphics.py

# default startup options
WIDTH, HEIGHT = 800, 600
fullscreen = False
N = 40 # lawn size

import sys
while len(sys.argv) > 1 and sys.argv[1].startswith('-'):
    if sys.argv[1] == '-f' or sys.argv[1] == '--fullscreen': # or toggle at runtime with key.F
        fullscreen = True
    elif (sys.argv[1] == '-N' or sys.argv[1] == '--lawnsize') and len(sys.argv) > 2:
        N = int(sys.argv[2])
        del sys.argv[2]
    else:
        print "unknown option", sys.argv[1]
    del sys.argv[1]
    continue

window_options = dict(fullscreen = fullscreen,
                      resizable = True,
                      width = WIDTH, height = HEIGHT )
if fullscreen:
    # Width and height cannot be specified with fullscreen.
    del window_options['width']
    del window_options['height']
    
try:
    # Try and create a window with multisampling (antialiasing)
    config = Config(sample_buffers = 1, samples = 4, 
                    depth_size = 16, double_buffer = True,)
    window = pyglet.window.Window(config = config, **window_options)
    ## print "using multisampling (antialiasing)" # this works for me
except pyglet.window.NoSuchConfigException:
    # Fall back to no multisampling for old hardware
    window = pyglet.window.Window(**window_options)
    print "note: NOT using multisampling (antialiasing)"

keyboard = key.KeyStateHandler()
window.push_handlers(keyboard)

class game_handlers(object):

    def on_resize(self, width, height):
        # Override the default on_resize handler to create a 3D projection
        glViewport(0, 0, width, height)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(60., width / float(height), .1, 1000.)
        glMatrixMode(GL_MODELVIEW)
        return pyglet.event.EVENT_HANDLED

    def on_draw(self):
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glLoadIdentity()
        window.camera.applyTransform()
        
##            print "using fallback viewpoint code (bug?)"
##            glTranslatef(0, 0, -4)
##            glRotatef(rz, 0, 0, 1)
##            glRotatef(ry, 0, 1, 0)
##            glRotatef(rx, 1, 0, 0)
##            glScalef(zoom, zoom, zoom) #bks - probably not the right place -- should be just viewpoint, not model
            
        batch.draw()
        
        # the following causes nothing to happen or be drawn, and cpu to go even higher --
        # maybe i need to change it in events too??
        ## window.invalid = False ### re issue 399, guess at how to put it in, evidently not needed on all macs only some

    def on_text(self, text): #bks, works
        global zoom, rotating
        if text == 'r':
            rotating = not rotating
            return pyglet.event.EVENT_HANDLED
        if text == ',' or text == '<':
            zoom *= 1.05**(-1)
            return pyglet.event.EVENT_HANDLED
        if text == '.' or text == '>':
            zoom *= 1.05**(1)
            return pyglet.event.EVENT_HANDLED

    def on_mouse_drag(self, x, y, dx, dy, button, modifiers): #bks
        global rotating, rx, ry, rz
        if modifiers & key.MOD_OPTION:
            rotating = False
            if modifiers & key.MOD_SHIFT:
                # pan
                offset_x += dx # wrong and nim, will traceback since vars not defined
                offset_y += dy
            else:
                # trackball
                rx += dx
                ry += dy # guesses, wrong
        return pyglet.event.EVENT_HANDLED

    def on_mouse_scroll(self, x, y, dx, dy): #bks
        # this comment is so I also find this if I search for on_mouse_wheel
        global zoom
        zoom *= 1.05**(dy) # move top of wheel same direction as camera should move (towards or away from you)

    def on_key_press(self, symbol, modifiers):

        if symbol == key.F:
            window.set_fullscreen(not window._fullscreen)
            return pyglet.event.EVENT_HANDLED

        if symbol == key.F8:
            world.terrain.debug_toggle_wireframe()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.SPACE:
            for p in world.players(): # main player, buddies, and enemies (all moving objects)
                p.control_jump() # enemies ignore this
            return pyglet.event.EVENT_HANDLED

        if symbol == key.N:
            world.player.control_add_buddy()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.DELETE or symbol == key.BACKSPACE: # on Mac, key.BACKSPACE seems to be the DELETE key
            world.control_delete_buddy()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.Y: # F9 is taken by Mac OS; F7 may have logged me out w/o saving changes (not sure); used to be E
            for p in world.players():
                p.debug_toggle_collisions()
            return pyglet.event.EVENT_HANDLED
        
        if symbol == key.E:
            world.player.control_add_enemy()
            return pyglet.event.EVENT_HANDLED

        if symbol == key.P:
            for p in world.players():
                p.pause_and_print_player_pos()
            return pyglet.event.EVENT_HANDLED

        if 0:
            # enable this to find key symbols
            print "key press ignored:", symbol
            for x in dir(key):
                if getattr(key, x) == symbol:
                    print " (%r)" % x


window.push_handlers(game_handlers())



class Camera(object):
    def __init__(self, player):
        self.player = player # XXX and world for up?
    def applyTransform(self):
        global zoom
        self.player.lookat(zoom)

def update(dt):
    if rotating:
        global rx, ry, rz
        rx += dt * 1
        ry += dt * 80
        rz += dt * 30
        rx %= 360
        ry %= 360
        rz %= 360
    world.update(dt)

pyglet.clock.schedule(update)
# I tried the following on a mac whose fan turns on from this game, but it didn't help and maybe hurt.
# no time to fully understand this or optimize it. (less than 3 hours til end of pyweek!)
## pyglet.clock.schedule_interval(update, 1/60.) #### was schedule, no time arg. check whether I'm calling the right function.
    # bks added time interval... (btw if i add one, do I then need window.invalid = False in on_draw re issue 399? no, that's separate)


def gl_setup():
    # One-time GL setup
    glClearColor(1, 1, 1, 1)
    ## glColor3f(1, 0, 0)
    glColor3f(0.0, 0.5, 0.5) # this change has no visible effect when glMaterialfv is used, or when it's not used;
        # works when lighting disabled; with glColorMaterial enabled, it affects terrain, but not (I'm pretty sure) the sphere.
        # Note that sphere has bright specularity when small but not when large (presumably due to various bugs in lighting & color).
    
    glEnable(GL_DEPTH_TEST)
##    glEnable(GL_CULL_FACE) ### why does it still sometimes disappear? turns out it was the specularity, causing complete whiteouts.

    # Uncomment this line for a wireframe view
    #glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)

    # Simple light setup.  On Windows GL_LIGHT0 is enabled by default,
    # but this is not the case on Linux or Mac, so remember to always 
    # include it.

##    glEnable(GL_LIGHTING)
    # if I disable lighting here, then I see my colored sphere, and glColor3f above (flat) affects terrain.
    # (general OpenGL lighting/color question: how do I get the effect of *both* lighting and per-vertex coloring?
    #  (while using Batch, not glMaterialfv)

    # hmm, maybe this tells me how: http://www.sjbaker.org/steve/omniv/opengl_lighting.html
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE)
        # this says glColor or vertex colors set (these two kinds of) material colors for lighting
    glEnable(GL_COLOR_MATERIAL)
    glEnable(GL_LIGHTING)

    
    glEnable(GL_LIGHT0)
##    glEnable(GL_LIGHT1)

    # Define a simple function to create ctypes arrays of floats:
    def vec(*args):
        return (GLfloat * len(args))(*args)

    glLightfv(GL_LIGHT0, GL_POSITION, vec(.5, .5, 1, 0))
##    glLightfv(GL_LIGHT0, GL_SPECULAR, vec(.5, .5, 1, 1))
    glLightfv(GL_LIGHT0, GL_DIFFUSE, vec(1, 1, 1, 1))
##    glLightfv(GL_LIGHT1, GL_POSITION, vec(1, 0, .5, 0))
##    glLightfv(GL_LIGHT1, GL_DIFFUSE, vec(.5, .5, .5, 1))
####    glLightfv(GL_LIGHT1, GL_SPECULAR, vec(1, 1, 1, 1))

    # see if zapping this helps show colors:
##    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, vec(0.5, 0, 0.3, 1))
    
##    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, vec(1, 1, 1, 1))
##    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 50)

# ==

world = player = None

batch = None

def main():
##    print "Hello from your game's main()"
##    print data.load('sample.txt').read()

    gl_setup()

    global batch, world
    
    batch = pyglet.graphics.Batch()
    world = World(N, batch = batch)
    world.keyboard = keyboard # for controlling the players (see world & player update methods)

    window.camera = Camera(world.player)
    
    pyglet.app.run()
