from pyglet.gl import *
import pyglet
from pyglet.window import key

from constants import * # kluge; some are overridden below

from terrain import Terrain

from player import Player, Buddy, Enemy

from euclid import Point3, Plane, Vector3
ORIGIN = Point3(0,0,0) # also in constants.py with float coords
VZERO = Vector3(0.,0.,0.)
UP = Vector3(0.,1.,0.) # also in constants.py 

from math import pi, sin, cos


def clamp(x, x1, x2): # warning: differs from the other def of this
    # note: if range is inconsistent, x1 (minumum allowed)
    # takes precedence over x2 (max allowed)
    return max(x1, min(x2, x))

class PlayerGroup(pyglet.graphics.Group): # XX USE ME
    "a single group to hold all players"
    def set_state(self):
        glEnable(GL_CULL_FACE)
    def unset_state(self):
        glDisable(GL_CULL_FACE)
    pass

class World:
    keyboard = None # should be set externally (kluge) to a KeyStateHandler
    _players = _enemies = ()
    
    def __init__(self, N, batch, group = None):
        self.terrain = Terrain( ORIGIN, Point3(1,0,0), Point3(0,0,1), N, N, noise = 0.1)
        self.terrain.add_to_batch(batch, group)

        self.player = Player()
        self.player.add_to_batch(batch, group)
        self._players = [self.player]
        self._enemies = []
        self.batch = batch
        self.group = group
        self.playergroup = PlayerGroup(self.group)
##        if 1:
##            # just to see if it works (controls might affect both of them for now!
##            # some do, some don't, depending on implem -- e.g. motion does, jump doesn't, key.Y toggle doesn't.)
##            # (and the players have no interaction for now)
##            self.player2 = Player(Point3(20,6,20))
##            self.player2.add_to_batch(batch, group)

    def players(self):
        return list(self._players) + list(self._enemies)

    def control_delete_buddy(self):
        if len(self._players) > 1:
            old = self._players[-1]
            del self._players[-1]
            old.remove_from_batch()
        
    def inter_player_force_pair(self, p1, p2):
        ## return VZERO, VZERO + UP * GRAVITY * 0.9 # testing stub
        # note: kluge: if it helps, you can assume p1 comes before p2 in self.players().
        # we don't use this yet except in the testing stub commented out above.
        force_onset_distance_2 = (p1.collision_radius + p2.collision_radius) ** 2
        p1_to_p2_vec = p2.pos - p1.pos
        actual_distance_2 = p1_to_p2_vec.magnitude_squared()
        if actual_distance_2 < force_onset_distance_2:
            pseudodepth = (force_onset_distance_2 - actual_distance_2)
            if pseudodepth > 0.5:
                pseudodepth = 0.5
            force = (pseudodepth/0.2) ** 3 * GRAVITY # pushing force between players
            p1_to_p2_direction = p1_to_p2_vec.normalized()
            f2 = p1_to_p2_direction * force # p1 pushes on p2
            f1 = - f2 # and vice versa
            return f1, f2
        else:
            return VZERO, VZERO
    
    def update(self, dt):
        if dt == 0.0:
            ## print "dt is 0!"
            # this happens once when we start;
            # intercept it to prevent division by 0
            return
        
        # first compute inter-player force pairs (usually equal and opposite)
        # (see comment in player verlet update method for why we do this separately)

        force_from_others = {} # player -> force on it from other players
        if not self.player.paused: # kluge
            interplayer_force_pairs = {}
                # maps (p1, p2) -> (f1, f2), where f1 = force on p1 from p2;
            def store(p1, p2, f1, f2):
                # we store two entries at once, to prevent redundant computations
                interplayer_force_pairs[p1, p2] = (f1, f2)
                interplayer_force_pairs[p2, p1] = (f2, f1)

            for p1 in self.players():
                for p2 in self.players():
                    if p2 is not p1 and not interplayer_force_pairs.has_key((p1, p2)):
                        f1, f2 = self.inter_player_force_pair(p1, p2)
                        store(p1, p2, f1, f2)

            for p1 in self.players():
                f1tot = Vector3(0.,0.,0.) # not VZERO, since we modify it
                for p2 in self.players():
                    if p2 is not p1:
                        f1, f2 = interplayer_force_pairs[p1, p2]
                        f1tot += f1
                force_from_others[p1] = f1tot
        else:
            for p in self.players():
                force_from_others[p] = VZERO

        # now update player states & positions, and perform player side effects on lawn

        for player in self.players():

            what_to_do = player.update(dt,
                                       self.terrain,
                                       self.keyboard,
                                       force_from_others = force_from_others[player] )
            if what_to_do is None:
                what_to_do = {}

            make_hill = what_to_do.get('make_hill', False)
            make_depression = what_to_do.get('make_depression', False)
            recolor = what_to_do.get('recolor', False) # false or new color
            add_buddy = what_to_do.get('add_buddy', False) # false or its initial position
            add_enemy = what_to_do.get('add_enemy', False) # false or its initial position

            if what_to_do.get('I_fell_out', False): # only enemies can do this
                try:
                    self._enemies.remove(player)
                except:
                    # should never happen, but be safe (last minute kluge, sorry)
                    pass 
                else:
                    player.remove_from_batch()
            
            if recolor:
                # for debugging, or for fun, recolor the under-player quad
                self.terrain.recolor_vertices_near(player.pos, 0, recolor) # radius = 0: only one quad
            
            if make_hill:
                def heightfunc(old, dx, dz): # note: uses old only for clamp limits
                    goal = max(0., (9. - (dx*dx+dz*dz))) * 0.3
                    rate = dt * 2. * goal/3.
                    return clamp( old + rate, old, goal )
                self.terrain.raise_vertices_near( player.pos, 5., heightfunc)

            if make_depression:
                # bugs: not symmetric with make_hill (maybe the 0.?); doesn't undo it either (need rate proportional to goal?)
                def heightfunc(old, dx, dz):
                    goal = max(0., (9. - (dx*dx+dz*dz))) * 0.3
                    rate = dt * 2. * goal/3.
                    return - clamp( - old + rate, - old, goal )
                self.terrain.raise_vertices_near( player.pos, 5., heightfunc)

            if add_buddy:
                new = Buddy(add_buddy, paused = self.player.paused)
                batch = self.batch
                group = self.playergroup
                new.add_to_batch(batch, group)
                self._players.append(new)

            if add_enemy: ### change how this is done - controlled by level, when you change levels; for now, diff key
                new = Enemy(add_enemy, paused = self.player.paused)
                batch = self.batch
                group = self.playergroup
                new.add_to_batch(batch, group)
                self._enemies.append(new)
            

    def start(self): ### XXX probably this belongs elsewhere 
        pyglet.clock.schedule(self.update)

    pass
